using UnityEngine;
using UnityEngine.UI;
using UnityEngine.Events;
using System;
using System.Text;
using System.Collections.Generic;
using MVR.FileManagementSecure;
using SimpleJSON;

namespace MacGruber
{
	public partial class IdlePoser : MVRScript
	{
		private List<JSONStorableParam> myMenuStorables = new List<JSONStorableParam>();
		private List<UIDynamic> myMenuElements = new List<UIDynamic>();

		private JSONStorableUrl myDataFile;
		private JSONStorableStringChooser myMainMenu;
		private JSONStorableStringChooser myMainState;
		private JSONStorableString myGeneralInfo;
		private JSONStorableString myPlayInfo;
		private JSONStorableStringChooser myCapturesControlList;
		private JSONStorableBool myStateAutoTransition;
		private JSONStorableStringChooser myAnchorCaptureList;
		private JSONStorableStringChooser myAnchorModeList;
		private JSONStorableStringChooser myAnchorAtomListA;
		private JSONStorableStringChooser myAnchorAtomListB;
		private JSONStorableStringChooser myAnchorControlListA;
		private JSONStorableStringChooser myAnchorControlListB;
		private JSONStorableFloat myAnchorBlendRatio;
		private JSONStorableFloat myAnchorDampingTime;
		private JSONStorableStringChooser myTransitionList;
		private JSONStorableBool myDebugShowInfo;
		private JSONStorableBool myDebugShowCurves;

		private readonly List<string> myStateTypes = new List<string>() {
			"Regular State",
			COLORTAG_CONTROLPOINT+"Control Point</color>",
			COLORTAG_INTERMEDIATE+"Intermediate Point</color>"
		};
		private readonly List<string> myStateGroups = new List<string>() {
			"None", "Group A", "Group B", "Group C", "Group D",
			"Group E", "Group F", "Group G", "Group H"
		};
		private readonly List<string> myAnchorModes = new List<string>() {
			"World",
			"Single Anchor",
			"Blend Anchor"
		};

		private const string FILE_EXTENSION = "idlepose";
		private const string BASE_DIRECTORY = "Custom/IdlePoses";
		private const string COLORTAG_INTERMEDIATE = "<color=#804b00>";
		private const string COLORTAG_CONTROLPOINT = "<color=#005780>";

		private GameObject myLabelWithXButtonPrefab;

		// =======================================================================================


		private void InitUI()
		{
			myDataFile = new JSONStorableUrl("IdlePose", "", UILoadJSON, FILE_EXTENSION, true);

			List<string> menuItems = new List<string>() {
				"Play", "Captures", "States", "Transitions", "Anchors", "Debug"
			};
			myMainMenu = new JSONStorableStringChooser("Menu", menuItems, menuItems[0], "Menu");
			myMainMenu.setCallbackFunction += UISelectMenu;

			List<string> stateItems = new List<string>();
			myMainState = new JSONStorableStringChooser("State", stateItems, "", "State");
			myMainState.setCallbackFunction += UISelectStateAndRefresh;

			myGeneralInfo = new JSONStorableString("Info", "");
			myPlayInfo = new JSONStorableString("Info", "");
			
			pluginLabelJSON.setCallbackFunction += (string label) => {
				if (!string.IsNullOrEmpty(label))
					label = ": " + label;
				myGeneralInfo.val = "<color=#606060><size=40><b>IdlePoser"+label+"</b></size></color>\n" +				
					"Character idle animations are stored as a network of <i>States</i> with <i>Transitions</i> between them. Use special <i>Control</i> and <i>Intermediate Point</i> states for smooth transition animations using Bézier curves.\n\n" +
					"Use the <i>Menu</i> dropdown to switch between the different UI pages.";
			};
			pluginLabelJSON.setCallbackFunction(pluginLabelJSON.val);

			myCapturesControlList = new JSONStorableStringChooser("Control", new List<string>(), "", "Control");
			myStateAutoTransition = new JSONStorableBool("Auto-Transition on State Change", true);
			myTransitionList = new JSONStorableStringChooser("Transition", new List<string>(), "", "Transition");
			myAnchorCaptureList = new JSONStorableStringChooser("ControlCapture", new List<string>(), "", "Control Capture");
			myAnchorCaptureList.setCallbackFunction += (string v) => UIRefreshMenu();

			myAnchorModeList = new JSONStorableStringChooser("AnchorMode", myAnchorModes, "", "Anchor Mode");
			myAnchorModeList.setCallbackFunction += (string v) => UISetAnchors();
			myAnchorAtomListA = new JSONStorableStringChooser("AnchorAtomA", new List<string>(), "", "Anchor Atom A");
			myAnchorAtomListA.setCallbackFunction += (string v) => UISetAnchors();
			myAnchorAtomListB = new JSONStorableStringChooser("AnchorAtomB", new List<string>(), "", "Anchor Atom B");
			myAnchorAtomListB.setCallbackFunction += (string v) => UISetAnchors();
			myAnchorControlListA = new JSONStorableStringChooser("AnchorControlA", new List<string>(), "", "Anchor Control A");
			myAnchorControlListA.setCallbackFunction += (string v) => UISetAnchors();
			myAnchorControlListB = new JSONStorableStringChooser("AnchorControlB", new List<string>(), "", "Anchor Control B");
			myAnchorControlListB.setCallbackFunction += (string v) => UISetAnchors();
			myAnchorBlendRatio = new JSONStorableFloat("Anchor Blend Ratio", 0.5f, 0.0f, 1.0f, true, true);
			myAnchorBlendRatio.setCallbackFunction += (float v) => UISetAnchorsBlendDamp();
			myAnchorDampingTime = new JSONStorableFloat("Anchor Damping Time", 0.2f, 0.0f, 5.0f, true, true);
			myAnchorDampingTime.setCallbackFunction += (float v) => UISetAnchorsBlendDamp();

			myDebugShowInfo = new JSONStorableBool("Show Debug Info", false);
			myDebugShowInfo.setCallbackFunction += (bool v) => UIRefreshMenu();
			myDebugShowCurves = new JSONStorableBool("Draw Curves in 3D View", false);
			myDebugShowCurves.setCallbackFunction += DebugSwitchShowCurves;

			CreateMainUI();
			UIRefreshMenu();
		}

		private void OnDestroyUI()
		{
			if (myLabelWithXButtonPrefab != null)
				Destroy(myLabelWithXButtonPrefab);
			myLabelWithXButtonPrefab = null;

			DestroyDebugCurves();
		}

		private void InitPrefabs()
		{
			myLabelWithXButtonPrefab = new GameObject("LabelXButton");
			myLabelWithXButtonPrefab.SetActive(false);
			RectTransform rt = myLabelWithXButtonPrefab.AddComponent<RectTransform>();
			rt.anchorMax = new Vector2(0, 1);
			rt.anchorMin = new Vector2(0, 1);
			rt.offsetMax = new Vector2(535, -500);
			rt.offsetMin = new Vector2(10, -600);
			LayoutElement le = myLabelWithXButtonPrefab.AddComponent<LayoutElement>();
			le.flexibleWidth = 1;
			le.minHeight = 50;
        	le.minWidth = 350;
        	le.preferredHeight = 50;
        	le.preferredWidth = 500;

			RectTransform backgroundTransform = manager.configurableScrollablePopupPrefab.transform.Find("Background") as RectTransform;
			backgroundTransform = Instantiate(backgroundTransform, myLabelWithXButtonPrefab.transform);
			backgroundTransform.name = "Background";
			backgroundTransform.anchorMax = new Vector2(1, 1);
			backgroundTransform.anchorMin = new Vector2(0, 0);
			backgroundTransform.offsetMax = new Vector2(0, 0);
			backgroundTransform.offsetMin = new Vector2(0, -10);

			RectTransform buttonTransform = manager.configurableScrollablePopupPrefab.transform.Find("Button") as RectTransform;
			buttonTransform = Instantiate(buttonTransform, myLabelWithXButtonPrefab.transform);
			buttonTransform.name = "Button";
			buttonTransform.anchorMax = new Vector2(1, 1);
			buttonTransform.anchorMin = new Vector2(1, 0);
			buttonTransform.offsetMax = new Vector2(0, 0);
			buttonTransform.offsetMin = new Vector2(-60, -10);
			Button buttonButton = buttonTransform.GetComponent<Button>();
			Text buttonText = buttonTransform.Find("Text").GetComponent<Text>();
			buttonText.text = "X";

			RectTransform labelTransform = buttonText.rectTransform;
			labelTransform = Instantiate(labelTransform, myLabelWithXButtonPrefab.transform);
			labelTransform.name = "Text";
			labelTransform.anchorMax = new Vector2(1, 1);
			labelTransform.anchorMin = new Vector2(0, 0);
			labelTransform.offsetMax = new Vector2(-65, 0);
			labelTransform.offsetMin = new Vector2(5, -10);
			Text labelText = labelTransform.GetComponent<Text>();
			labelText.verticalOverflow = VerticalWrapMode.Overflow;

			UIDynamicLabelXButton uid = myLabelWithXButtonPrefab.AddComponent<UIDynamicLabelXButton>();
			uid.label = labelText;
			uid.button = buttonButton;
		}

		private void UIRefreshMenu()
		{
			myMainMenu.setCallbackFunction(myMainMenu.val);
		}

		private void UISelectMenu(string menuItem)
		{
			myPaused = true;
			myMainState.popup.visible = true; // Workaround for PopupPanel appearing behind other UI for some reason
			myMainState.popup.visible = false;
			CleanupMenu();

			List<string> states = new List<string>();
			foreach (var state in myStates)
				states.Add(state.Key);
			states.Sort();
			List<string> stateDisplays = new List<string>(states.Count);
			for (int i=0; i<states.Count; ++i)
			{
				State state = myStates[states[i]];
				if (state.IsIntermediate)
					stateDisplays.Add(COLORTAG_INTERMEDIATE+state.myName+"</color>");
				else if (state.IsControlPoint)
					stateDisplays.Add(COLORTAG_CONTROLPOINT+state.myName+"</color>");
				else
					stateDisplays.Add(state.myName);
			}
			myMainState.choices = states;
			myMainState.displayChoices = stateDisplays;
			if (states.Count == 0)
			{
				myMainState.valNoCallback = "";
			}
			else if (!states.Contains(myMainState.val))
			{
				myMainState.valNoCallback = states[0];
				UIBlendToState();
			}

			if (menuItem == "Play")
				CreatePlayMenu();
			else if (menuItem == "Captures")
				CreateCapturesMenu();
			else if (menuItem == "States")
				CreateStateMenu();
			else if (menuItem == "Anchors")
				CreateAnchorsMenu();
			else if (menuItem == "Transitions")
				CreateTransitionsMenu();
			else if (menuItem == "Debug")
				CreateDebugMenu();
		}

		private void UIBlendToState()
		{
			State state;
			if (myStateAutoTransition.val && myStates.TryGetValue(myMainState.val, out state))
				SetBlendTransition(state);
		}

		private void UISelectStateAndRefresh(string name)
		{
			if (myMainMenu.val != "Play")
				UIBlendToState();
			UIRefreshMenu();
		}

		private void CreateMainUI()
		{
			CreateMenuInfo(myGeneralInfo, 344, true);

			{
				UIDynamicButton button = CreateButton("Load", false);
				myDataFile.setCallbackFunction -= UILoadJSON;
				myDataFile.allowFullComputerBrowse = false;
				myDataFile.allowBrowseAboveSuggestedPath = false;
				myDataFile.shortCuts = FileManagerSecure.GetShortCutsForDirectory(
					BASE_DIRECTORY, allowNavigationAboveRegularDirectories: false,
					useFullPaths: false, generateAllFlattenedShortcut: true,
					includeRegularDirsInFlattenedShortcut: true);
				myDataFile.SetFilePath(BASE_DIRECTORY+"/");
				myDataFile.RegisterFileBrowseButton(button.button);
				myDataFile.setCallbackFunction += UILoadJSON;

				CreateMenuButton("Save As", UISaveJSONDialog, false);
			}

			CreateMenuPopup(myMainMenu, false);
			CreateMenuPopup(myMainState, false);			

			CreateMenuSyncSpacer(15);

			myMenuStorables.Clear();
			myMenuElements.Clear();
		}

		private void CreatePlayMenu()
		{
			CreateMenuInfoOneLine("<size=30><b>Play Idle</b></size>", false);

			if (!myDebugShowInfo.val)
			{
				if (myStates.Count > 0)
					myPlayInfo.val = "IdlePoser is playing animations.";
				else
					myPlayInfo.val = "You need to add some states and transitions before you can play animations.";
			}

			CreateMenuInfo(myPlayInfo, 200, false);

			myPaused = false;
		}

		private void CreateCapturesMenu()
		{
			// control captures
			CreateMenuInfoOneLine("<size=30><b>Control Captures</b></size>", false);

			FreeControllerV3[] atomControls = GetAllControllers();
			List<string> availableControls = new List<string>();
			for (int i=0; i<atomControls.Length; ++i)
			{
				string control = atomControls[i].name;
				if (control.StartsWith("hair"))
					continue;
				if (myControlCaptures.FindIndex(cc => cc.myName == control) >= 0)
					continue;
				availableControls.Add(control);
			}
			myCapturesControlList.choices = availableControls;
			if (availableControls.Count == 0)
				myCapturesControlList.val = "";
			else if (!availableControls.Contains(myCapturesControlList.val))
				myCapturesControlList.val = availableControls[0];
			CreateMenuPopup(myCapturesControlList, false);
			CreateMenuButton("Add Controller", UIAddControlCapture, false);

			if (myControlCaptures.Count > 0)
				CreateMenuSpacer(15, false);

			for (int i=0; i<myControlCaptures.Count; ++i)
			{
				ControlCapture cc = myControlCaptures[i];
				CreateMenuLabelXButton(cc.myName, () => UIRemoveControlCapture(cc), false);
			}

			// morph captures
			CreateMenuInfoOneLine("<size=30><b>Morph Captures</b></size>", true);

			CreateMenuButton("Add Animated Morphs", UIAddMorphCapture, true);

			if (myMorphCaptures.Count > 0)
				CreateMenuSpacer(15, false);

			for (int i=0; i<myMorphCaptures.Count; ++i)
			{
				MorphCapture mc = myMorphCaptures[i];
				DAZMorph m = mc.myMorph;
				if (m.hasBoneModificationFormulas)
				{
					// this is a "high cost" morph...red warning
					CreateMenuLabelXButton(
						"<color=#CC6060>[" +mc.myGender + "] " + m.resolvedRegionName + "</color><color=#FF0000>\n" + m.resolvedDisplayName + "</color>",
						() => UIRemoveMorphCapture(mc), true);
				}
				else
				{
					CreateMenuLabelXButton(
						"<color=#606060>[" +mc.myGender + "] " + m.resolvedRegionName + "</color>\n" + m.resolvedDisplayName,
						() => UIRemoveMorphCapture(mc), true);
				}
			}
		}

		private void CreateStateMenu()
		{
			CreateMenuInfoOneLine("<size=30><b>State Manager</b></size>", false);
			CreateMenuSyncSpacer(0);
			CreateMenuButton("Add State", UIAddState, false);
			State state;
			if (!myStates.TryGetValue(myMainState.val, out state))
				return;

			CreateMenuButton("Remove State", UIRemoveState, false);
			CreateMenuButton("Capture State", UICaptureState, true);
			CreateMenuToggle(myStateAutoTransition, true);

			CreateMenuSpacer(15, false);

			CreateMenuInfoOneLine("<size=30><b>State Settings</b></size>", false);

			CreateMenuSyncSpacer(0);

			JSONStorableString name = new JSONStorableString("State Name", state.myName, UIRenameState);
			CreateMenuTextInput(name, false);



			JSONStorableStringChooser typeChooser = new JSONStorableStringChooser("StateType", myStateTypes, myStateTypes[state.myStateType], "State\nType");
			typeChooser.setCallbackFunction += (string v) => {
				State s = UIGetState();
				if (s != null)
					s.myStateType = myStateTypes.IndexOf(v);
				UIRefreshMenu();
			};
			CreateMenuPopup(typeChooser, false);

			if (state.IsRegularState)
			{
				JSONStorableStringChooser groupChooser = new JSONStorableStringChooser("StateGroup", myStateGroups, myStateGroups[state.myStateGroup], "State\nGroup");
				groupChooser.setCallbackFunction += (string v) => {
					State s = UIGetState();
					if (s != null)
						s.myStateGroup = myStateGroups.IndexOf(v);
				};
				CreateMenuPopup(groupChooser, false);
			}

			JSONStorableFloat probability = new JSONStorableFloat("Relative Probability", state.myProbability, 0.0f, 1.0f, true, true);
			probability.setCallbackFunction = (float v) => {
				State s = UIGetState();
				if (s != null)
					s.myProbability = v;
			};
			CreateMenuSlider(probability, true);

			JSONStorableFloat transitionDuration = new JSONStorableFloat("Transition Duration", state.myTransitionDuration, 0.01f, 5.0f, true, true);
			transitionDuration.setCallbackFunction = (float v) => {
				State s = UIGetState();
				if (s != null)
					s.myTransitionDuration = v;
			};
			CreateMenuSlider(transitionDuration, true);

			JSONStorableFloat waitDurationMin = new JSONStorableFloat("Wait Duration Min", state.myWaitDurationMin, 0.0f, 300.0f, true, true);
			waitDurationMin.setCallbackFunction = (float v) => {
				State s = UIGetState();
				if (s != null)
					s.myWaitDurationMin = v;
			};
			CreateMenuSlider(waitDurationMin, true);

			JSONStorableFloat waitDurationMax = new JSONStorableFloat("Wait Duration Max", state.myWaitDurationMax, 0.0f, 300.0f, true, true);
			waitDurationMax.setCallbackFunction = (float v) => {
				State s = UIGetState();
				if (s != null)
					s.myWaitDurationMax = v;
			};
			CreateMenuSlider(waitDurationMax, true);
		}

		private void CreateAnchorsMenu()
		{
			CreateMenuInfoOneLine("<size=30><b>Anchors</b></size>", false);

			State state;
			if (!myStates.TryGetValue(myMainState.val, out state))
			{
				CreateMenuInfo("You need to add a state before you can configure anchors.", 100, false);
				return;
			}

			List<string> captures = new List<string>(myControlCaptures.Count);
			for (int i=0; i<myControlCaptures.Count; ++i)
				captures.Add(myControlCaptures[i].myName);
			myAnchorCaptureList.choices = captures;
			if (captures.Count == 0)
				myAnchorCaptureList.valNoCallback = "";
			else if (!captures.Contains(myAnchorCaptureList.val))
				myAnchorCaptureList.valNoCallback = captures[0];
			CreateMenuPopup(myAnchorCaptureList, false);

			ControlCapture controlCapture = myControlCaptures.Find(cc => cc.myName == myAnchorCaptureList.val);
			if (controlCapture == null)
				return;

			CreateMenuSpacer(30, false);

			ControlEntryAnchored controlEntry;
			if (!state.myControlEntries.TryGetValue(controlCapture, out controlEntry))
				return;
			myAnchorModeList.valNoCallback = myAnchorModes[controlEntry.myAnchorMode];
			CreateMenuPopup(myAnchorModeList, false);

			if (controlEntry.myAnchorMode != ControlEntryAnchored.ANCHORMODE_WORLD)
			{
				bool isBlendMode = controlEntry.myAnchorMode == ControlEntryAnchored.ANCHORMODE_BLEND;

				List<string> visibleAtoms = GetAtomUIDs();
				List<string> atoms = new List<string>(visibleAtoms.Count+1);
				atoms.Add("[Self]");
				atoms.AddRange(visibleAtoms);
				myAnchorAtomListA.label = isBlendMode ? "Anchor Atom A" : "Anchor Atom";
				myAnchorAtomListA.valNoCallback = controlEntry.myAnchorAAtom;
				myAnchorAtomListA.choices = atoms;
				CreateMenuPopup(myAnchorAtomListA, false);

				Atom atomA = GetSelfOrAtomById(myAnchorAtomListA.val);
				List<string> controls = atomA?.GetStorableIDs() ?? new List<string>();
				controls.RemoveAll((string control) => {
					if (atomA == containingAtom && control == myAnchorCaptureList.val)
						return true;
					JSONStorable storable = atomA.GetStorableByID(control);
					if (storable is DAZBone || storable is FreeControllerV3)
						return false;
					return true;
				});
				myAnchorControlListA.label = isBlendMode ? "Anchor Control A" : "Anchor Control";
				myAnchorControlListA.valNoCallback = controlEntry.myAnchorAControl;
				myAnchorControlListA.choices = controls;
				if (controls.Count == 0)
					myAnchorControlListA.valNoCallback = "";
				else if (!controls.Contains(myAnchorControlListA.val))
					myAnchorControlListA.valNoCallback = controls[0];
				CreateMenuPopup(myAnchorControlListA, false);

				if (isBlendMode)
				{
					myAnchorAtomListB.valNoCallback = controlEntry.myAnchorBAtom;
					myAnchorAtomListB.choices = atoms;
					CreateMenuPopup(myAnchorAtomListB, false);

					Atom atomB = GetSelfOrAtomById(myAnchorAtomListB.val);
					controls = atomB?.GetStorableIDs() ?? new List<string>();
					controls.RemoveAll((string control) => {
						if (atomB == containingAtom && control == myAnchorCaptureList.val)
							return true;
						JSONStorable storable = atomB.GetStorableByID(control);
						if (storable is DAZBone || storable is FreeControllerV3)
							return false;
						return true;
					});
					myAnchorControlListB.valNoCallback = controlEntry.myAnchorBControl;
					myAnchorControlListB.choices = controls;
					if (controls.Count == 0)
						myAnchorControlListB.valNoCallback = "";
					else if (!controls.Contains(myAnchorControlListB.val))
						myAnchorControlListB.valNoCallback = controls[0];
					CreateMenuPopup(myAnchorControlListB, false);

					myAnchorBlendRatio.valNoCallback = controlEntry.myBlendRatio;
					CreateMenuSlider(myAnchorBlendRatio, false);
				}

				myAnchorDampingTime.valNoCallback = controlEntry.myDampingTime;
				CreateMenuSlider(myAnchorDampingTime, false);
			}
		}

		private void CreateTransitionsMenu()
		{
			CreateMenuInfoOneLine("<size=30><b>Transitions</b></size>", false);

			State state;
			if (!myStates.TryGetValue(myMainState.val, out state))
			{
				CreateMenuInfo("You need to add some states before you can add transitions.", 100, false);
				return;
			}

			List<State> transitions = new List<State>(state.myTransitions);
			transitions.Sort((State a, State b) => a.myName.CompareTo(b.myName));
			for (int i=0; i<transitions.Count; ++i)
			{
				string label;
				State target = transitions[i];
				if (target.IsIntermediate)
					label = COLORTAG_INTERMEDIATE+target.myName+"</color>";
				else if (target.IsControlPoint)
					label = COLORTAG_CONTROLPOINT+target.myName+"</color>";
				else
					label = target.myName;

				CreateMenuLabelXButton(label, () => UIRemoveTransition(state, target), false);
			}
			if (state.myTransitions.Count == 0)
			{
				CreateMenuInfoOneLine("There are no transitions for this state.", false);
			}

			CreateMenuSpacer(15, false);

			List<string> availableTargets = new List<string>();
			foreach (var t in myStates)
			{
				State target = t.Value;
				if (state == target)
					continue;
				if (state.myTransitions.Contains(target))
					continue;
				availableTargets.Add(target.myName);
			}
			availableTargets.Sort();
			List<string> availableTargetDisplays = new List<string>(availableTargets.Count);
			for (int i=0; i<availableTargets.Count; ++i)
			{
				State target = myStates[availableTargets[i]];
				if (target.IsIntermediate)
					availableTargetDisplays.Add(COLORTAG_INTERMEDIATE+target.myName+"</color>");
				else if (target.IsControlPoint)
					availableTargetDisplays.Add(COLORTAG_CONTROLPOINT+target.myName+"</color>");
				else
					availableTargetDisplays.Add(target.myName);
			}
			myTransitionList.choices = availableTargets;
			myTransitionList.displayChoices = availableTargetDisplays;
			if (availableTargets.Count == 0)
				myTransitionList.val = "";
			else if (!availableTargets.Contains(myTransitionList.val))
				myTransitionList.val = availableTargets[0];
			CreateMenuPopup(myTransitionList, false);
			CreateMenuButton("Add Transition", UIAddTransition, false);
		}

		private void CreateDebugMenu()
		{
			CreateMenuInfoOneLine("<size=30><b>Debug Options</b></size>", false);
			CreateMenuInfoOneLine("<color=#ff0000><b>Can cause performance issues!</b></color>", false);
						
			if (myDebugShowInfo.val)
				CreateMenuInfo(myPlayInfo, 200, false);

			CreateMenuToggle(myDebugShowInfo, false);
			CreateMenuToggle(myDebugShowCurves, false);
		}


		// =======================================================================================


		private void UILoadJSON(string url)
		{		
			JSONClass jc = LoadJSON(url).AsObject;
			if (jc != null)
				LoadPose(jc);			
			
			if (myCurrentState != null)
			{
				myMainState.valNoCallback = myCurrentState.myName;
				myMainState.setCallbackFunction(myCurrentState.myName);
			}
			else
			{
				UIRefreshMenu();
			}
		}

		private void UISaveJSONDialog()
		{
			SuperController sc = SuperController.singleton;
			sc.GetMediaPathDialog(UISaveJSON, FILE_EXTENSION, BASE_DIRECTORY, false, true, false, null, false, null, false, false);
			sc.mediaFileBrowserUI.SetTextEntry(true);
			if (sc.mediaFileBrowserUI.fileEntryField != null)
			{
				string filename = ((int)(DateTime.UtcNow - new DateTime(1970, 1, 1)).TotalSeconds).ToString();
				sc.mediaFileBrowserUI.fileEntryField.text = filename + "." + FILE_EXTENSION;
				sc.mediaFileBrowserUI.ActivateFileNameField();
			}
		}

		private void UISaveJSON(string path)
		{
			if (string.IsNullOrEmpty(path))
				return;
			path = path.Replace('\\', '/');
			SuperController.LogMessage("IdlePoser: Saving as '"+path+"'.");
			JSONClass jc = SavePose();
			SaveJSON(jc, path);
		}

		private void UIAddControlCapture()
		{
			ControlCapture cc = new ControlCapture(this, myCapturesControlList.val);
			myControlCaptures.Add(cc);
			foreach (var state in myStates)
				cc.CaptureEntry(state.Value);
			UIRefreshMenu();
		}

		private void UIRemoveControlCapture(ControlCapture cc)
		{
			myControlCaptures.Remove(cc);
			foreach (var state in myStates)
				state.Value.myControlEntries.Remove(cc);
			UIRefreshMenu();
		}

		private void UIAddMorphCapture()
		{
			DAZCharacterSelector geometry = containingAtom.GetStorableByID("geometry") as DAZCharacterSelector;
			if (geometry == null)
				return;

			bool isMale = geometry.gender == DAZCharacterSelector.Gender.Male;
			if (!isMale || geometry.GetBoolParamValue("useFemaleMorphsOnMale"))
				UIAddMorphCapture(DAZCharacterSelector.Gender.Female, geometry.morphsControlFemaleUI);
			if (isMale || geometry.GetBoolParamValue("useMaleMorphsOnFemale"))
				UIAddMorphCapture(DAZCharacterSelector.Gender.Male, geometry.morphsControlMaleUI);

			UIRefreshMenu();
		}

		private void UIAddMorphCapture(DAZCharacterSelector.Gender gender, GenerateDAZMorphsControlUI morphsControl)
		{
			List<DAZMorph> morphs = morphsControl.GetMorphs();
			for (int i=0; i<morphs.Count; ++i)
			{
				DAZMorph morph = morphs[i];
				if (!morph.animatable)
					continue;
				if (myMorphCaptures.FindIndex(x => x.myMorph == morph && x.myGender == gender) >= 0)
					continue;

				MorphCapture mc = new MorphCapture(gender, morph);
				myMorphCaptures.Add(mc);
				foreach (var state in myStates)
					mc.CaptureEntry(state.Value);
			}
		}

		private void UIRemoveMorphCapture(MorphCapture mc)
		{
			myMorphCaptures.Remove(mc);
			foreach (var state in myStates)
				state.Value.myMorphEntries.Remove(mc);
			UIRefreshMenu();
		}

		private void UIAddState()
		{
			for (int i=1; i<1000; ++i)
			{
				string name = "State#" + i;
				if (!myStates.ContainsKey(name))
				{
					CreateState(name);
					myMainState.val = name;
					UIRefreshMenu();
					return;
				}
			}
			SuperController.LogError("IdlePoser: Too many states!");
		}

		private void UICaptureState()
		{
			State state = UIGetState();
			if (state == null)
				return;

			CaptureState(state);
			UIRefreshMenu();
		}

		private void UIRemoveState()
		{
			State state = UIGetState();
			if (state == null)
				return;

			myStates.Remove(state.myName);
			foreach (var s in myStates)
				s.Value.myTransitions.RemoveAll(x => x == state);

			UIRefreshMenu();
		}

		private void UIRenameState(string name)
		{
			State state = UIGetState();
			if (state == null)
				return;
			if (state.myName == name)
				return;

			myStates.Remove(state.myName);

			int altIndex = 2;
			string baseName = name;
			while (myStates.ContainsKey(name))
			{
				name = baseName + "#" + altIndex;
				++altIndex;
			}

			myStates.Add(name, state);
			state.myName = name;			
			myMainState.valNoCallback = name;
			UIRefreshMenu();
		}

		private void UISetAnchors()
		{
			State state = UIGetState();
			if (state == null)
				return;

			ControlCapture controlCapture = myControlCaptures.Find(cc => cc.myName == myAnchorCaptureList.val);
			if (controlCapture == null)
				return;

			ControlEntryAnchored controlEntry;
			if (!state.myControlEntries.TryGetValue(controlCapture, out controlEntry))
				return;

			int anchorMode = myAnchorModes.FindIndex(m => m == myAnchorModeList.val);
			if (anchorMode >= 0)
				controlEntry.myAnchorMode = anchorMode;
			controlEntry.myBlendRatio = myAnchorBlendRatio.val;
			controlEntry.myDampingTime = myAnchorDampingTime.val;

			if (anchorMode >= ControlEntryAnchored.ANCHORMODE_SINGLE)
			{
				controlEntry.myAnchorAAtom = myAnchorAtomListA.val;
				controlEntry.myAnchorAControl = myAnchorControlListA.val;
			}
			if (anchorMode == ControlEntryAnchored.ANCHORMODE_BLEND)
			{
				controlEntry.myAnchorBAtom = myAnchorAtomListB.val;
				controlEntry.myAnchorBControl = myAnchorControlListB.val;
			}
			controlEntry.AdjustAnchor(this);

			UIRefreshMenu();
		}
		
		private void UISetAnchorsBlendDamp()
		{
			State state = UIGetState();
			if (state == null)
				return;

			ControlCapture controlCapture = myControlCaptures.Find(cc => cc.myName == myAnchorCaptureList.val);
			if (controlCapture == null)
				return;

			ControlEntryAnchored controlEntry;
			if (!state.myControlEntries.TryGetValue(controlCapture, out controlEntry))
				return;

			controlEntry.myBlendRatio = myAnchorBlendRatio.val;
			controlEntry.myDampingTime = myAnchorDampingTime.val;
			controlEntry.AdjustAnchor(this);
		}

		private void UIAddTransition()
		{
			State source = UIGetState();
			if (source == null)
				return;
			State target;
			if (!myStates.TryGetValue(myTransitionList.val, out target))
				return;

			if (!source.myTransitions.Contains(target))
				source.myTransitions.Add(target);
			if (!target.myTransitions.Contains(source))
				target.myTransitions.Add(source);

			UIRefreshMenu();
		}

		private void UIRemoveTransition(State source, State target)
		{
			source.myTransitions.Remove(target);
			target.myTransitions.Remove(source);
			UIRefreshMenu();
		}

		private State UIGetState()
		{
			State state;
			if (!myStates.TryGetValue(myMainState.val, out state))
			{
				SuperController.LogError("IdlePoser: Invalid state selected!");
				return null;
			}
			else
			{
				return state;
			}
		}

		// =======================================================================================

		/*private static void LogTransform(string message, Transform t)
		{
			StringBuilder b = new StringBuilder();
			b.Append(message).Append("\n");
			LogTransformInternal(t, 0, b);
			SuperController.LogMessage(b.ToString());
		}

		private static void LogTransformInternal(Transform t, int indent, StringBuilder b)
		{
			b.Append(' ', indent*4).Append(t.name).Append("\n");

			Component[] comps = t.GetComponents<Component>();
			if (comps.Length > 0)
			{
				b.Append(' ', indent*4+2).Append("Components:\n");
				for (int i=0; i<comps.Length; ++i)
				{
					Component c = comps[i];
					b.Append(' ', indent*4+4).Append(c.GetType().FullName).Append("\n");

					if (c is RectTransform)
					{
						RectTransform rt = c as RectTransform;
						b.Append(' ', indent*4+8).Append("anchoredPosition=").Append(rt.anchoredPosition).Append("\n");
						b.Append(' ', indent*4+8).Append("anchorMax=").Append(rt.anchorMax).Append("\n");
						b.Append(' ', indent*4+8).Append("anchorMin=").Append(rt.anchorMin).Append("\n");
						b.Append(' ', indent*4+8).Append("offsetMax=").Append(rt.offsetMax).Append("\n");
						b.Append(' ', indent*4+8).Append("offsetMin=").Append(rt.offsetMin).Append("\n");
						b.Append(' ', indent*4+8).Append("pivot=").Append(rt.pivot).Append("\n");
						b.Append(' ', indent*4+8).Append("rect=").Append(rt.rect).Append("\n");
					}
					else if (c is LayoutElement)
					{
						LayoutElement le = c as LayoutElement;
						b.Append(' ', indent*4+8).Append("flexibleHeight=").Append(le.flexibleHeight).Append("\n");
						b.Append(' ', indent*4+8).Append("flexibleWidth=").Append(le.flexibleWidth).Append("\n");
						b.Append(' ', indent*4+8).Append("ignoreLayout=").Append(le.ignoreLayout).Append("\n");
						b.Append(' ', indent*4+8).Append("layoutPriority=").Append(le.layoutPriority).Append("\n");
						b.Append(' ', indent*4+8).Append("minHeight=").Append(le.minHeight).Append("\n");
						b.Append(' ', indent*4+8).Append("minWidth=").Append(le.minWidth).Append("\n");
						b.Append(' ', indent*4+8).Append("preferredHeight=").Append(le.preferredHeight).Append("\n");
						b.Append(' ', indent*4+8).Append("preferredWidth=").Append(le.preferredWidth).Append("\n");
					}
					else if (c is Image)
					{
						Image img = c as Image;
						b.Append(' ', indent*4+8).Append("mainTexture=").Append(img.mainTexture.name).Append("\n");
						b.Append(' ', indent*4+8).Append("sprite=").Append(img.sprite.name).Append("\n");
					}
				}
			}
			if (t.childCount > 0)
			{
				b.Append(' ', indent*4+2).Append("Children:\n");
				for (int i=0; i<t.childCount; ++i)
				{
					Transform c = t.GetChild(i);
					LogTransformInternal(c, indent+1, b);
				}
			}
		}*/


		// =======================================================================================


		private void CreateMenuButton(string label, UnityAction callback, bool rightSide)
		{
			UIDynamicButton button = CreateButton(label, rightSide);
			button.button.onClick.AddListener(callback);
			myMenuElements.Add(button);
		}

		private void CreateMenuLabelXButton(string label, UnityAction callback, bool rightSide)
		{
			if (myLabelWithXButtonPrefab == null)
				InitPrefabs();

			Transform t = CreateUIElement(myLabelWithXButtonPrefab.transform, rightSide);
			UIDynamicLabelXButton uid = t.gameObject.GetComponent<UIDynamicLabelXButton>();
			uid.label.text = label;
			uid.button.onClick.AddListener(callback);
			myMenuElements.Add(uid);
			t.gameObject.SetActive(true);
		}

		private void CreateMenuToggle(JSONStorableBool storable, bool rightSide)
		{
			CreateToggle(storable, rightSide);
			myMenuStorables.Add(storable);
		}

		private void CreateMenuSlider(JSONStorableFloat storable, bool rightSide)
		{
			CreateSlider(storable, rightSide);
			myMenuStorables.Add(storable);
		}

		private void CreateMenuPopup(JSONStorableStringChooser storable, bool rightSide)
		{
			CreateScrollablePopup(storable, rightSide);
			myMenuStorables.Add(storable);
		}

		private void CreateMenuTextInput(JSONStorableString storable, bool rightSide)
        {
            UIDynamicTextField textfield = CreateTextField(storable, rightSide);
            textfield.height = 35;
            textfield.backgroundColor = Color.white;
			LayoutElement layout = textfield.GetComponent<LayoutElement>();
			layout.preferredHeight = layout.minHeight = 35;
            InputField inputfield = textfield.gameObject.AddComponent<InputField>();
            inputfield.GetComponent<RectTransform>().sizeDelta = new Vector2(1f, 0.4f);
            inputfield.textComponent = textfield.UItext;
            storable.inputField = inputfield;
			myMenuStorables.Add(storable);
        }

		private void CreateMenuInfo(string text, float height, bool rightSide)
		{
			JSONStorableString storable = new JSONStorableString("Info", text);
			UIDynamicTextField textfield = CreateTextField(storable, rightSide);
			textfield.height = height;
			myMenuStorables.Add(storable);
		}

		private void CreateMenuInfo(JSONStorableString storable, float height, bool rightSide)
		{
			UIDynamicTextField textfield = CreateTextField(storable, rightSide);
			textfield.height = height;
			myMenuStorables.Add(storable);
		}

		private void CreateMenuInfoOneLine(string text, bool rightSide)
		{
			JSONStorableString storable = new JSONStorableString("Info", text);
			UIDynamicTextField textfield = CreateTextField(storable, rightSide);
			textfield.UItext.alignment = TextAnchor.LowerLeft;
			LayoutElement layout = textfield.GetComponent<LayoutElement>();
			layout.preferredHeight = layout.minHeight = 35;
			textfield.height = 35;
			myMenuStorables.Add(storable);
		}

		private void CreateMenuSpacer(float height, bool rightSide)
		{
			UIDynamic spacer = CreateSpacer(rightSide);
			spacer.height = height;
			myMenuElements.Add(spacer);
		}

		private void CreateMenuSyncSpacer(float height)
		{
			float SPACING = 15;
			float leftHeight = 0.0f;
			if (leftUIElements.Count >= 1)
				leftHeight += (leftUIElements.Count-1) * SPACING;
			for (int i=0; i < leftUIElements.Count; ++i)
			{
				RectTransform rt = leftUIElements[i] as RectTransform;
				if (rt != null)
					leftHeight += rt.sizeDelta.y;
			}
			float rightHeight = 0.0f;
			if (rightUIElements.Count >= 1)
				rightHeight += (rightUIElements.Count-1) * SPACING;
			for (int i=0; i < rightUIElements.Count; ++i)
			{
				RectTransform rt = rightUIElements[i] as RectTransform;
				if (rt != null)
					rightHeight += rt.sizeDelta.y;
			}

			height = Mathf.Max(height, 0.0f);
			bool extraHeight = height > 0;
			bool side = leftHeight > rightHeight;
			UIDynamic spacer = CreateSpacer(side);
			spacer.height = Mathf.Abs(leftHeight - rightHeight) + (extraHeight ? height : -SPACING);
			myMenuElements.Add(spacer);
			if (extraHeight)
			{
				spacer = CreateSpacer(!side);
				spacer.height = height;
				myMenuElements.Add(spacer);
			}
		}

		private void CleanupMenu()
		{
			for (int i=0; i<myMenuStorables.Count; ++i)
			{
				JSONStorableParam jsp = myMenuStorables[i];
				if (jsp is JSONStorableFloat)
					RemoveSlider(jsp as JSONStorableFloat);
				else if (jsp is JSONStorableBool)
					RemoveToggle(jsp as JSONStorableBool);
				else if (jsp is JSONStorableColor)
					 RemoveColorPicker(jsp as JSONStorableColor);
				else if (jsp is JSONStorableString)
					RemoveTextField(jsp as JSONStorableString);
				else if (jsp is JSONStorableStringChooser)
				{
					// Workaround for VaM not cleaning its panels properly.
					JSONStorableStringChooser jssc = jsp as JSONStorableStringChooser;
					RectTransform popupPanel = jssc.popup?.popupPanel;
					RemovePopup(jssc);
					if (popupPanel != null)
						UnityEngine.Object.Destroy(popupPanel.gameObject);
				}
			}
			myMenuStorables.Clear();

			for (int i=0; i<myMenuElements.Count; ++i)
			{
				UIDynamic uid = myMenuElements[i];
				if (uid is UIDynamicButton)
					RemoveButton(uid as UIDynamicButton);
				else if (uid is UIDynamicLabelXButton)
					RemoveSpacer(uid);
				else if (uid is UIDynamicSlider)
					RemoveSlider(uid as UIDynamicSlider);
				else if (uid is UIDynamicToggle)
					RemoveToggle(uid as UIDynamicToggle);
				else if (uid is UIDynamicColorPicker)
					 RemoveColorPicker(uid as UIDynamicColorPicker);
				else if (uid is UIDynamicTextField)
					RemoveTextField(uid as UIDynamicTextField);
				else if (uid is UIDynamicPopup)
				{
					// Workaround for VaM not cleaning its panels properly.
					UIDynamicPopup uidp = uid as UIDynamicPopup;
					RectTransform popupPanel = uidp.popup?.popupPanel;
					RemovePopup(uidp);
					if (popupPanel != null)
						UnityEngine.Object.Destroy(popupPanel.gameObject);
				}
				else
					RemoveSpacer(uid);
			}
			myMenuElements.Clear();
		}


		// =======================================================================================


		private class UIDynamicLabelXButton : UIDynamic
		{
			public Text label;
			public Button button;
		}

		// =======================================================================================

		private StringBuilder myPlayInfoBuilder;

		private bool myDebugCurvesActive = false;

		private Mesh myDebugLineMesh;
		private Material myDebugLineMaterial;
		private List<Vector3> myDebugLineVertices;
		private List<int> myDebugLineIndices;
		private int[] myDebugLineIndicesBuffer;
		private List<State> myDebugTransition;
		private HashSet<ulong> myDebugTransitionHashes;
		private Vector3[] myDebugLinePositions;
		private Vector3[] myDebugLineTempA;
		private Vector3[] myDebugLineTempB;

		private GameObject myDebugCubePrefab;
		private Material[] myDebugStateCubeMaterials;
		private List<GameObject>[] myDebugStateCubes = new List<GameObject>[3];
		private int[] myDebugStateCubesCounts = new int[3];
		private const float DEBUG_CUBE_SIZE = 0.007f;

		private void DebugSwitchShowCurves(bool enabled)
		{
			if (enabled && !myDebugCurvesActive)
				InitDebugCurves();
			else if (!enabled && myDebugCurvesActive)
				DestroyDebugCurves();
		}

		private void InitDebugCurves()
		{
			Shader[] shaders = Resources.FindObjectsOfTypeAll<Shader>();
			Shader shader = Array.Find(shaders, s => s.name == "Custom/TransparentHUD");

			// lines
			myDebugLineMaterial = new Material(shader);
			myDebugLineMaterial.color = Color.blue;
			myDebugLineMesh = new Mesh();
			myDebugLineMesh.MarkDynamic();

			myDebugLineVertices = new List<Vector3>();
			myDebugLineIndices = new List<int>();
			myDebugTransition = new List<State>(MAX_STATES);
			myDebugTransitionHashes = new HashSet<ulong>();
			myDebugLinePositions = new Vector3[MAX_STATES];
			myDebugLineTempA = new Vector3[MAX_STATES-1];
			myDebugLineTempB = new Vector3[MAX_STATES-2];

			// cubes
			myDebugCubePrefab = GameObject.CreatePrimitive(PrimitiveType.Cube);
			myDebugCubePrefab.SetActive(false);
			myDebugCubePrefab.transform.parent = containingAtom.transform;
			myDebugCubePrefab.transform.localScale = new Vector3(DEBUG_CUBE_SIZE, DEBUG_CUBE_SIZE, DEBUG_CUBE_SIZE);
			DestroyImmediate(myDebugCubePrefab.GetComponent<BoxCollider>());

			myDebugStateCubeMaterials = new Material[NUM_STATETYPES];
			Material m = new Material(shader);
			m.color = Color.blue;
			myDebugStateCubeMaterials[STATETYPE_REGULARSTATE] = m;
			m = new Material(shader);
			m.color = new Color(0/255.0f, 87/255.0f, 128/255.0f);
			myDebugStateCubeMaterials[STATETYPE_CONTROLPOINT] = m;
			m = new Material(shader);
			m.color = new Color(128/255.0f, 75/255.0f, 0/255.0f);
			myDebugStateCubeMaterials[STATETYPE_INTERMEDIATE] = m;

			myDebugStateCubes = new List<GameObject>[NUM_STATETYPES];
			for (int i=0; i<NUM_STATETYPES; ++i)
				myDebugStateCubes[i] = new List<GameObject>();
			myDebugStateCubesCounts = new int[NUM_STATETYPES];

			myDebugCurvesActive = true;
		}

		private void DestroyDebugCurves()
		{
			if (!myDebugCurvesActive)
				return;
			myDebugCurvesActive = false;

			Destroy(myDebugLineMesh);
			Destroy(myDebugLineMaterial);
			myDebugLineMesh = null;
			myDebugLineMaterial = null;
			myDebugLineVertices = null;
			myDebugLineIndices = null;
			myDebugLineIndicesBuffer = null;
			myDebugTransition = null;
			myDebugLinePositions = null;
			myDebugLineTempA = null;
			myDebugLineTempB = null;


			Destroy(myDebugCubePrefab);
			myDebugCubePrefab = null;

			for (int i=0; i<NUM_STATETYPES; ++i)
			{
				List<GameObject> cubes = myDebugStateCubes[i];
				for (int j=0; j<cubes.Count; ++j)
					Destroy(cubes[j]);
				cubes.Clear();
				Destroy(myDebugStateCubeMaterials[i]);
			}
			myDebugStateCubeMaterials = null;
			myDebugStateCubes = null;
			myDebugStateCubesCounts = null;
		}

		private void DebugUpdateUI()
		{
			if (myDebugCurvesActive)
			{
				if (!myPaused || myNextState != null || myTransition.Count > 0)
				{
					foreach (var s in myStates)
					{
						State state = s.Value;
						if (state == myCurrentState || myCurrentTransition.Contains(state))
							continue; // these have been already updated
						for (int i=0; i<myControlCaptures.Count; ++i)
							myControlCaptures[i].UpdateState(state);
					}
				}
				
				DebugUpdateCubes();
				DebugUpdateLines();
			}

			if (myDebugShowInfo.val)
			{
				if (myPlayInfoBuilder == null)
					myPlayInfoBuilder = new StringBuilder();
				else
					myPlayInfoBuilder.Length = 0;

				if (myNextState != null)
				{
					myPlayInfoBuilder.Append("Current Transition:\n  ");
					myPlayInfoBuilder.AppendFormat("{0:N2}", myClock).Append(" / ").AppendFormat("{0:N2}", myDuration);
					myPlayInfoBuilder.Append("\n  ").Append(myCurrentTransition[0].myName);
					for (int i=1; i<myCurrentTransition.Count; ++i)
						myPlayInfoBuilder.Append("\n  => ").Append(myCurrentTransition[i].myName);
				}
				else
				{
					myPlayInfoBuilder.Append("Current State:\n  ");
					myPlayInfoBuilder.AppendFormat("{0:N2}", myClock).Append(" / ").AppendFormat("{0:N2}", myDuration);
					myPlayInfoBuilder.Append("\n  ");
					if (myCurrentState != null)
						myPlayInfoBuilder.Append(myCurrentState.myName);
					else
						myPlayInfoBuilder.Append("NULL");
				}

				myPlayInfo.val = myPlayInfoBuilder.ToString();
			}
		}

		private void DebugUpdateCubes()
		{
			for (int i=0; i<NUM_STATETYPES; ++i)
				myDebugStateCubesCounts[i] = 0;

			// update cube positions
			foreach (var s in myStates)
			{
				State state = s.Value;
				int stateType = state.myStateType;
				if (stateType >= NUM_STATETYPES)
					continue;
				List<GameObject> cubes = myDebugStateCubes[stateType];
				foreach (var e in state.myControlEntries)
				{
					ControlEntryAnchored ce = e.Value;
					int cubeIndex = myDebugStateCubesCounts[stateType]++;

					GameObject cube;
					if (cubeIndex >= cubes.Count)
					{
						cube = Instantiate(myDebugCubePrefab, containingAtom.transform);
						cube.GetComponent<Renderer>().material = myDebugStateCubeMaterials[stateType];
						cube.SetActive(true);
						cubes.Add(cube);
					}
					else
					{
						cube = cubes[cubeIndex];
					}
					ce.myEntry.SetTransform(cube.transform);
				}
			}

			// cleanup unused cubes
			for (int i=0; i<NUM_STATETYPES; ++i)
			{
				int cubeCount = myDebugStateCubesCounts[i];
				List<GameObject> cubes = myDebugStateCubes[i];
				if (cubeCount < cubes.Count)
				{
					for (int j=cubeCount; j<cubes.Count; ++j)
						Destroy(cubes[j]);
					cubes.RemoveRange(cubeCount, cubes.Count-cubeCount);
				}
			}
		}

		private void DebugUpdateLines()
		{
			myDebugLineVertices.Clear();
			myDebugLineIndices.Clear();
			myDebugTransitionHashes.Clear();
			uint index = 0;
			foreach (var s in myStates)
				s.Value.myDebugIndex = index++;
				
			foreach (var s in myStates)
			{
				State state = s.Value;
				if (state.IsControlPoint)
					continue;
				myDebugTransition.Add(state);
				DebugGatherTransitionLines();
				myDebugTransition.Clear();
			}

			if (myDebugLineVertices.Count == 0)
			{
				myDebugLineMesh.Clear();
			}
			else
			{
				if (myDebugLineIndicesBuffer == null || myDebugLineIndicesBuffer.Length != myDebugLineIndices.Count)
					myDebugLineIndicesBuffer = new int[myDebugLineIndices.Count];
				for (int i=0; i<myDebugLineIndices.Count; ++i)
					myDebugLineIndicesBuffer[i] = myDebugLineIndices[i];
				myDebugLineMesh.SetVertices(myDebugLineVertices);
				myDebugLineMesh.SetIndices(myDebugLineIndicesBuffer, MeshTopology.Lines, 0);

				Graphics.DrawMesh(myDebugLineMesh, Matrix4x4.identity, myDebugLineMaterial, gameObject.layer, null, 0, null, castShadows: false, receiveShadows: false);
			}
		}

		private void DebugGatherTransitionLines()
		{
			State source = myDebugTransition[0];
			State current = myDebugTransition[myDebugTransition.Count-1];
			for (int i=0; i<current.myTransitions.Count; ++i)
			{
				State next = current.myTransitions[i];
				if (myDebugTransition.Contains(next))
					continue;

				if (next.IsControlPoint)
				{
					if (myDebugTransition.Count >= MAX_STATES-1)
						continue;
					myDebugTransition.Add(next);
					DebugGatherTransitionLines();
					myDebugTransition.RemoveAt(myDebugTransition.Count-1);
				}
				else
				{
					if (next.IsRegularState)
					{
						if (next.myProbability < 0.01f)
							continue;
						if (source.myStateGroup != 0 && source.myStateGroup == next.myStateGroup)
							continue;
					}

					myDebugTransition.Add(next);
					DebugGatherTransitionLinesSample();
					myDebugTransition.RemoveAt(myDebugTransition.Count-1);
				}
			}
		}
		
		private bool DebugHasSampledTransition()
		{
			ulong hash = 0;
			int entryCount = myDebugTransition.Count;
			if (myDebugTransition[0].myDebugIndex < myDebugTransition[entryCount-1].myDebugIndex)
			{
				for (int i=0; i<entryCount; ++i)
					hash = (hash << 10) | (myDebugTransition[i].myDebugIndex & 0x3FF);
			}
			else
			{
				for (int i=entryCount-1; i>=0; --i)
					hash = (hash << 10) | (myDebugTransition[i].myDebugIndex & 0x3FF);
			}
			
			return !myDebugTransitionHashes.Add(hash);
		}

		private void DebugGatherTransitionLinesSample()
		{
			if (DebugHasSampledTransition())
				return;
			
			int entryCount = myDebugTransition.Count;
			int numSamples = Mathf.Max(entryCount-2, 0) * 8;
			int numLines = numSamples + 1;
			for (int i=0; i<myControlCaptures.Count; ++i)
			{
				ControlCapture cc = myControlCaptures[i];
				for (int j=0; j<entryCount; ++j)
				{
					ControlEntryAnchored ce;
					if (!myDebugTransition[j].myControlEntries.TryGetValue(cc, out ce))
						continue;
					myDebugLinePositions[j] = ce.myEntry.myPosition;
				}

				int startVertex = myDebugLineVertices.Count;
				myDebugLineVertices.Add(myDebugLinePositions[0]);
				for (int j=1; j<=numSamples; ++j)
				{
					float t = j / (float)(numSamples+1);

					// evaluating Bézier curve
					Vector3[] curr = myDebugLineTempA;
					Vector3[] prev = myDebugLineTempB;
					for (int k=0; k<entryCount-1; ++k)
					{
						curr[k] = Vector3.LerpUnclamped(myDebugLinePositions[k], myDebugLinePositions[k+1], t);
					}
					for (int level=1; level<entryCount-1; ++level)
					{
						Vector3[] tmp = curr;
						curr = prev;
						prev = tmp;
						for (int k=0; k<entryCount-level-1; ++k)
							curr[k] = Vector3.LerpUnclamped(prev[k], prev[k+1], t);
					}
					myDebugLineVertices.Add(curr[0]);
				}
				myDebugLineVertices.Add(myDebugLinePositions[entryCount-1]);

				int endVertex = startVertex+numLines;
				for (int j=startVertex; j<endVertex; ++j)
				{
					myDebugLineIndices.Add(j);
					myDebugLineIndices.Add(j+1);
				}
			}
		}
	}

}